using System;
using System.IO;
using System.Windows.Forms;
using GTA;
using GTA.Native;

public class IV_CE_OvertakeFix : Script
{
	//Turns Logging on or off
    private bool loggingEnabled = true;
    private bool displayTextEnabled = true; // use F4-Key in Game to toggle (or change further below)
	
	private readonly string logFilePath = Path.Combine(AppDomain.CurrentDomain.BaseDirectory, "IV_CE_OvertakeFix.log");
	
	//Prevents detection at the side of vehicle
	private const float TRUCK_DOT_PRODUCT_THRESHOLD = -0.96f;
    private const float CAR_DOT_PRODUCT_THRESHOLD = -0.94f;
    private const float BIKE_DOT_PRODUCT_THRESHOLD = -0.90f; 
	
	//How far away vehicle is detected
    private const float CAR_DETECTION_DISTANCE = 6.7f;
    private const float TRUCK_DETECTION_DISTANCE = 9.1f;
    private const float BIKE_DETECTION_DISTANCE = 3.6f;

	//Timer in milliseconds
    private const int FREEZE_DURATION = 14000; 

    // Truck state variables
    private Vehicle lastFrozenTruck;
    private DateTime lastTruckFrozenTime;
    private bool playerMovedSinceTruckUnfreeze = true;

    // Car state variables
    private Vehicle lastStoppedCar;
    private DateTime lastCarStoppedTime;
    private bool playerMovedSinceCarUnfreeze = true;

    // Bike state variables
    private Vehicle lastStoppedBike;
    private DateTime lastBikeStoppedTime;
    private bool playerMovedSinceBikeUnfreeze = true;
	
	// Flag to manage freezing behavior
	private bool canFreezeNextVehicle = true; 

    // Lists of model hashes for different vehicle types
    private readonly uint[] truckHashes = 
    {
		4011757448u,	// avan
		850991848u,		// biff
		3581397346u,	// bus
		2287941233u,	// pbus
		1938952078u,	// firetruk
		1353720154u,	// flatbed
		904750859u,		// mule
		569305213u,		// packer
		207497487u,		// packer2
		2157618379u,	// phantom
		1747439474u,	// stockade
		1911513875u,	// nstockade
		2394394458u,	// pstockade
		1917016601u,	// trash
		3195007082u,	// yankee
		2396846736u		// yankee2
    };

    private readonly uint[] bikeHashes = 
    {
		1672195559u,	// akuma
		3723957976u,	// angel
		4180675781u,	// bati
		3403504941u,	// bati2
		2464508460u,	// bobber
		2006142190u,	// daemon
		3886915065u,	// diabolus
		2623969160u,	// double
		2535109211u,	// double2
		2452219115u,	// faggio
		55628203u,		// faggio2
		1265391242u,	// hakuchou
		4039289119u,	// hakuchou2
		584879743u,		// hellfury
		301427732u,		// hexer
		4135840458u,	// innovation
		802082487u,		// lycan
		2688780135u,	// nightblade
		1203311498u,	// nrg900
		3385765638u,	// pcj
		4260343491u,	// policeb
		3935799761u,	// revenant
		788045382u,		// sanchez
		4154065143u,	// vader
		4217198264u,	// wayfarer
		3676349299u,	// wolfsbane
		3724934023u		// zombieb
	};

    public IV_CE_OvertakeFix()
    {
        Interval = 333;
        Tick += IV_CE_OvertakeFix_Tick;
        KeyDown += IV_CE_OvertakeFix_KeyDown;
        InitializeLogFile();
    }
    
    private void IV_CE_OvertakeFix_KeyDown(object sender, GTA.KeyEventArgs e)
    {
        if (e.Key == Keys.F4) //change desired Key here for IngameLogMessage
        {
            displayTextEnabled = !displayTextEnabled;
            Game.DisplayText("DisplayText " + (displayTextEnabled ? "Enabled" : "Disabled"));
        }
    }
    
    private void InitializeLogFile()
    {
        if (loggingEnabled && !File.Exists(logFilePath))
        {
            File.WriteAllText(logFilePath, "IV_CE_OvertakeFix Log File\n\n");
        }
    }

    private void LogMessage(string message)
    {
        if (loggingEnabled)
        {
            File.AppendAllText(logFilePath, DateTime.Now + ": " + message + Environment.NewLine);
        }
    }

    private void IV_CE_OvertakeFix_Tick(object sender, EventArgs e)
    {
        try
        {
            if (!Player.Character.isInVehicle()) return;

            Vehicle playerVeh = Player.Character.CurrentVehicle;
            if (playerVeh == null || !playerVeh.Exists()) return;

            bool playerStopped = playerVeh.Speed < 0.1f;

            // Reset freeze timers when the player moves
            if (!playerStopped)
            {
                playerMovedSinceTruckUnfreeze = true;
                playerMovedSinceCarUnfreeze = true;
                playerMovedSinceBikeUnfreeze = true;

                lastTruckFrozenTime = DateTime.Now;
                lastCarStoppedTime = DateTime.Now;
                lastBikeStoppedTime = DateTime.Now;

                if (!canFreezeNextVehicle)
                {
                    canFreezeNextVehicle = true;  // Allow freezing if the player starts moving again
                }
            }

            int recognitionDelay = playerStopped ? 60000 : 1; // 60 sec if stopped; else 1 ms

            Vehicle vehBehind = FindVehicleBehind(playerVeh, recognitionDelay);
            if (vehBehind != null)
            {
                LogAndStopVehicle(vehBehind, playerStopped);
            }
            else
            {
                UnfreezeVehiclesIfNeeded(playerStopped);
            }
        }
        catch (Exception ex)
        {
            LogMessage("Error in IV_CE_OvertakeFix: " + ex.Message);
        }
    }

    private Vehicle FindVehicleBehind(Vehicle playerVeh, int recognitionDelay)
    {
        Vehicle vehBehind = null;
        float minDistance = float.MaxValue;
        Vector3 playerPosition = playerVeh.Position;
        Vector3 playerDirection = playerVeh.Direction;

        foreach (Vehicle veh in World.GetAllVehicles())
        {
            if (IsValidVehicle(veh, playerVeh, recognitionDelay))
            {
                float distance = (veh.Position - playerPosition).Length();
                if (distance < minDistance && IsInDetectionRange(veh, playerVeh, distance, playerPosition, playerDirection))
                {
                    vehBehind = veh;
                    minDistance = distance;
                }
            }
        }

        return vehBehind;
    }

    private bool IsValidVehicle(Vehicle veh, Vehicle playerVeh, int recognitionDelay)
    {
        if (veh == playerVeh || !veh.Exists()) return false;
		
		LogMessage("Checking vehicle (Hash): " + veh.Model.Hash);

        if (IsTruck(veh.Model.Hash))
        {
            if (veh == lastFrozenTruck && (DateTime.Now - lastTruckFrozenTime).TotalMilliseconds < recognitionDelay)
                return false;
            if (!playerMovedSinceTruckUnfreeze) return false;
        }
        else if (IsBike(veh.Model.Hash))
        {
            if (veh == lastStoppedBike && (DateTime.Now - lastBikeStoppedTime).TotalMilliseconds < recognitionDelay)
                return false;
            if (!playerMovedSinceBikeUnfreeze) return false;
        }
        else // Treat as car
        {
            if (veh == lastStoppedCar && (DateTime.Now - lastCarStoppedTime).TotalMilliseconds < recognitionDelay)
                return false;
            if (!playerMovedSinceCarUnfreeze) return false;
        }
        return true;
    }

    private bool IsInDetectionRange(Vehicle veh, Vehicle playerVeh, float distance, Vector3 playerPosition, Vector3 playerDirection)
    {
        float maxDetectionDistance;
        if (IsTruck(veh.Model.Hash))
        {
            maxDetectionDistance = TRUCK_DETECTION_DISTANCE;
        }
        else if (IsBike(veh.Model.Hash))
        {
            maxDetectionDistance = BIKE_DETECTION_DISTANCE;
        }
        else
        {
            maxDetectionDistance = CAR_DETECTION_DISTANCE;
        }
        return distance <= maxDetectionDistance && IsInDirection(playerVeh, veh, playerPosition, playerDirection);
    }

    private bool IsInDirection(Vehicle playerVeh, Vehicle veh, Vector3 playerPosition, Vector3 playerDirection)
    {
        Vector3 toVehicle = veh.Position - playerPosition;
        float toVehicleMagnitudeSquared = toVehicle.X * toVehicle.X + toVehicle.Y * toVehicle.Y;
        if (toVehicleMagnitudeSquared < 0.0001f) return false;

        float dotProduct = (playerVeh.Direction.X * toVehicle.X + playerVeh.Direction.Y * toVehicle.Y) / (float)Math.Sqrt(toVehicleMagnitudeSquared);
        float threshold;
        if (IsTruck(veh.Model.Hash))
        {
            threshold = TRUCK_DOT_PRODUCT_THRESHOLD;
        }
        else if (IsBike(veh.Model.Hash))
        {
            threshold = BIKE_DOT_PRODUCT_THRESHOLD;
        }
        else
        {
            threshold = CAR_DOT_PRODUCT_THRESHOLD;
        }
        return dotProduct <= threshold;
    }
	
	private bool IsTruck(int modelHash)
    {
        foreach (uint hash in truckHashes)
        {
            if (hash == modelHash)
                return true;
        }
        return false;
    }

    private bool IsBike(int modelHash)
    {
		
        foreach (uint hash in bikeHashes)
        {
            if (hash == modelHash)
			{
                return true;
			}
        }
        
		return false;
    }

    private void LogAndStopVehicle(Vehicle vehBehind, bool playerStopped)
    {
        
        LogMessage("Detected vehicle behind (Hash): " + vehBehind.Model.Hash);

        Ped driver = vehBehind.GetPedOnSeat(VehicleSeat.Driver);
        if (driver != null && driver.Exists() && driver != Player.Character && playerStopped)
        {
            driver.Task.ClearAll();

            if (canFreezeNextVehicle)
            {
                if (IsTruck(vehBehind.Model.Hash))
				{
					if (displayTextEnabled)
					{
						Game.DisplayText("Detected Truck (Hash): " + vehBehind.Model.Hash);
					}
                	Function.Call("SET_CAR_FORWARD_SPEED", vehBehind, 0);
                    vehBehind.FreezePosition = true;
                    lastFrozenTruck = vehBehind;
                    lastTruckFrozenTime = DateTime.Now;
                    LogMessage("Stopped truck (Hash): " + vehBehind.Model.Hash);
                }
                else if (IsBike(vehBehind.Model.Hash)) // Properly handling bikes here
                {
					if (displayTextEnabled)
					{
						Game.DisplayText("Detected Bike (Hash): " + vehBehind.Model.Hash);
					}
                    Function.Call("SET_CAR_FORWARD_SPEED", vehBehind, 0);
                    driver.Task.StandStill(FREEZE_DURATION);
                    if (lastStoppedBike == null || !lastStoppedBike.Exists() || lastStoppedBike.Model.Hash != vehBehind.Model.Hash)
                    {
                        lastStoppedBike = vehBehind;
                        lastBikeStoppedTime = DateTime.Now;
                    }
                    LogMessage("Stopped bike (Hash): " + vehBehind.Model.Hash);
                }
                else
                {
                    // Car handling
					if (displayTextEnabled)
					{
						Game.DisplayText("Detected Car (Hash): " + vehBehind.Model.Hash);
					}
                    Function.Call("SET_CAR_FORWARD_SPEED", vehBehind, 0);
                    driver.Task.StandStill(FREEZE_DURATION);
                    if (lastStoppedCar == null || !lastStoppedCar.Exists() || lastStoppedCar.Model.Hash != vehBehind.Model.Hash)
                    {
                        lastStoppedCar = vehBehind;
                        lastCarStoppedTime = DateTime.Now;
                    }
                    LogMessage("Stopped car (Hash): " + vehBehind.Model.Hash);
                }
            }
        }
    }

    private void UnfreezeVehiclesIfNeeded(bool playerStopped)
    {
        if (lastFrozenTruck != null && lastFrozenTruck.Exists())
        {
            double truckFrozenDuration = (DateTime.Now - lastTruckFrozenTime).TotalMilliseconds;
            if (displayTextEnabled)
            {
                Game.DisplayText("Truck Frozen time elapsed: " + truckFrozenDuration + " ms");
            }
            if (!playerStopped || truckFrozenDuration >= FREEZE_DURATION)
            {
                int hash = lastFrozenTruck.Model.Hash;
                lastFrozenTruck.FreezePosition = false;
                Function.Call("SET_CAR_FORWARD_SPEED", lastFrozenTruck, 0);
                lastFrozenTruck = null;
                playerMovedSinceTruckUnfreeze = false;
                canFreezeNextVehicle = false;
                if (displayTextEnabled)
                {
                    Game.DisplayText("Unfreezing truck (Hash): " + hash);
                }
                LogMessage("Unfreezing truck (Hash): " + hash);
            }
        }

        if (lastStoppedCar != null && lastStoppedCar.Exists())
        {
            double carStoppedDuration = (DateTime.Now - lastCarStoppedTime).TotalMilliseconds;
            if (displayTextEnabled)
            {
                Game.DisplayText("Car Stopped time elapsed: " + carStoppedDuration + " ms");
            }
            if (!playerStopped || carStoppedDuration >= FREEZE_DURATION)
            {
                int hash = lastStoppedCar.Model.Hash;
                Ped driver = lastStoppedCar.GetPedOnSeat(VehicleSeat.Driver);
                if (driver != null && driver.Exists())
                {
                    driver.Task.ClearAll();
                    Function.Call("SET_CAR_FORWARD_SPEED", lastStoppedCar, 0);
                }
                lastStoppedCar = null;
                playerMovedSinceCarUnfreeze = false;
                canFreezeNextVehicle = false;
                if (displayTextEnabled)
                {
                    Game.DisplayText("Unfreezing car (Hash): " + hash);
                }
                LogMessage("Unfreezing car (Hash): " + hash);
            }
        }

        if (lastStoppedBike != null && lastStoppedBike.Exists())
        {
            double bikeStoppedDuration = (DateTime.Now - lastBikeStoppedTime).TotalMilliseconds;
            if (displayTextEnabled)
            {
                Game.DisplayText("Bike Stopped time elapsed: " + bikeStoppedDuration + " ms");
            }
            if (!playerStopped || bikeStoppedDuration >= FREEZE_DURATION)
            {
                int hash = lastStoppedBike.Model.Hash;
                Ped driver = lastStoppedBike.GetPedOnSeat(VehicleSeat.Driver);
                if (driver != null && driver.Exists())
                {
                    driver.Task.ClearAll();
                    Function.Call("SET_CAR_FORWARD_SPEED", lastStoppedBike, 0);
                }
                lastStoppedBike = null;
                playerMovedSinceBikeUnfreeze = false;
                canFreezeNextVehicle = false;
                if (displayTextEnabled)
                {
                    Game.DisplayText("Unfreezing bike (Hash): " + hash);
                }
                LogMessage("Unfreezing bike (Hash): " + hash);
            }
        }
    }
}
